#include <iostream>
#include <objbase.h>

#include "COMHelper.h"
#include "IFace.h"
#include "Registry.h"

using namespace std;

//Functions
void trace(char* msg) { cerr << msg; }

#ifdef DEBUG
void dbtrace(char* msg) { cerr << msg; }
#else
void dbtrace(char* msg {}
#endif

//-----------------

//Global variables
static HMODULE g_hModule = NULL;    //DLL handle
static long g_cComponents = 0;      //Components active
static long g_cServerLocks = 0;      //server locks

//Friendly name
const char g_szFrndName[] = "Jason's First COM";
//Version independent
const char g_szVerIndProgID[] = "Jason.FirstCOM";
//ProgID
const char g_szProgID[] = "Jason.FirstCOM.1";

//-------------------

class CJason : public IJason
{
public:
    //IUnknown
    virtual HRESULT __stdcall QueryInterface(const IID& iid, void** ppv);
    virtual ULONG __stdcall AddRef();
    virtual ULONG __stdcall Release();

    //IJason
    VVSTD fSaySomething() { trace("Something.\n"); }
    VVSTD fHelloCOMWorld() { trace("Hello COM World!\n"); }

    //constructor, destructor
    CJason();
    ~CJason();

private:
    //reference count
    long m_cRef;

};

//Constructor/destructor
CJason::CJason() : m_cRef(1)
{
dbtrace("In CJason().\n");
InterlockedIncrement(&g_cComponents);
}

CJason::~CJason()
{
dbtrace("In ~CJason().\n");
InterlockedDecrement(&g_cComponents);
}

//IUnkown functions
HRESULT __stdcall CJason::QueryInterface(const IID& iid, void** ppv)
{
if(iid == IID_IUnknown)
{
    *ppv = static_cast<IJason*>(this);
}

else if(iid == IID_IJason)
{
    *ppv = static_cast<IJason*>(this);
}

else
{
    *ppv = NULL;
    return E_NOINTERFACE;
}
reinterpret_cast<IUnknown*>(*ppv)->AddRef();
return S_OK;
}

ULONG __stdcall CJason::AddRef()
{
return InterlockedIncrement(&m_cRef);
}

ULONG __stdcall CJason::Release()
{
if(InterlockedDecrement(&m_cRef) == 0)
{
    delete this;
    return 0;
}
return m_cRef;
}

//Class factory

class CFactory : public IClassFactory
{
public:
    //IUnknown
    virtual HRESULT __stdcall QueryInterface(const IID& iid, void** ppv);
    virtual ULONG __stdcall AddRef();
    virtual ULONG __stdcall Release();

    //IClassFactory
    virtual HRESULT __stdcall CreateInstance(IUnknown* pUnknownOuter,
                                                const IID& iid, void** ppv);
    virtual HRESULT __stdcall LockServer(BOOL bLock);

    //Constructor, destructor
    CFactory() : m_cRef(1) { dbtrace("In CFactory().\n"); }
    ~CFactory() { dbtrace("In ~CFactory().\n"); }
    
private:
    long m_cRef;
};

//IUnknown
HRESULT __stdcall CFactory::QueryInterface(const IID& iid, void** ppv)
{
dbtrace("In CFactory::QueryInterface().\n");

if(iid == IID_IUnknown)
{
    dbtrace("Asked for IID_IUnknown.\n");
    *ppv = static_cast<IClassFactory*>(this);
}

if(iid == IID_IClassFactory)
{
    dbtrace("Asked for IID_IClassFactory.\n");
    *ppv = static_cast<IClassFactory*>(this);
}

else
{
    dbtrace("Asked for unsupported interface.\n");
    *ppv = NULL;
    return E_NOINTERFACE;
}

reinterpret_cast<IUnknown*>(*ppv)->AddRef();
return S_OK;
}

ULONG __stdcall CFactory::AddRef()
{
return InterlockedIncrement(&m_cRef);
}

ULONG __stdcall CFactory::Release()
{
if(InterlockedDecrement(&m_cRef) == 0)
{
    delete this;
    return 0;
}
return m_cRef;
}

//IClassFactory functions
HRESULT __stdcall CFactory::CreateInstance(IUnknown* pUnknownOuter,
                                            const IID& iid, void** ppv)
{
dbtrace("In CFactory::CreateInstance().\n");
if(pUnknownOuter != NULL)
{     
    return CLASS_E_NOAGGREGATION;
}

CJason* pJason = new CJason;
if(pJason == NULL)
{
    return E_OUTOFMEMORY;
}

//get interface
HRESULT hr = pJason->QueryInterface(iid, ppv);

//release IUnkown pointer, if QI failed component will delete
pJason->Release();
return hr;
}

HRESULT __stdcall CFactory::LockServer(BOOL bLock)
{
if(bLock)
{
    InterlockedIncrement(&g_cServerLocks);
}

else
{
    InterlockedDecrement(&g_cServerLocks);
}
return S_OK;
}

//DLL Functions

extern "C" __stdcall HRESULT DllCanUnloadNow()
{
if((g_cComponents == 0) && (g_cServerLocks == 0))
{
    return S_OK;
}
else
{
    return S_FALSE;
}
}

extern "C" __stdcall HRESULT DllGetClassObject(const CLSID& clsid, const IID& iid, void** ppv)
{
dbtrace("In DllGetClassObject().\n");
if(clsid != CLSID_CJason)
{
    dbtrace("clsid != CLSID_CJason.\n");
    return CLASS_E_CLASSNOTAVAILABLE;
}

trace("Getting CFactory.\n");
CFactory* pFactory = new CFactory; 
if(pFactory == NULL)
{
    return E_OUTOFMEMORY;
}

HRESULT hr = pFactory->QueryInterface(iid, ppv);
pFactory->Release();

return hr;
}

extern "C" __stdcall HRESULT DllRegisterServer()
{
return RegisterServer(g_hModule, CLSID_CJason, g_szFrndName, g_szVerIndProgID,
                    g_szProgID);
}

extern "C" __stdcall HRESULT DllUnregisterServer()
{
return UnregisterServer(CLSID_CJason, g_szVerIndProgID, g_szProgID);
}

extern "C"
BOOL APIENTRY DllMain(HANDLE hModule, DWORD dwReason, void* lpReserved)
{
switch(dwReason)
{
    case DLL_PROCESS_ATTACH:
        dbtrace("In DllMain(): DLL_PROCESS_ATTACH.\n");
        g_hModule = hModule;
        break;
    case DLL_PROCESS_DETACH:
        dbtrace("In DllMain(): DLL_PROCESS_DETACH.\n");
        break;
}

return TRUE;
}
